@@ -22,26 +22,28 @@ elif [[ $(git status --porcelain -b | grep -e "ahead" -e "behind") != "" ]]; the
 fi
 
 BRANCH="$(git rev-parse --abbrev-ref HEAD)"
-if [[ "$BRANCH" != "master" && "$BRANCH" != "main" ]]; then
-    read -p "You are pushing from a branch ('$BRANCH'), are you sure? " -n 1 -r
-    echo
-    if [[ $REPLY =~ ^[^Yy]$ ]]; then
-        echo Aborting.
-        exit 1
-    fi
-else
-  # This check is pretty naive, so it doesn't seem to work if you're releasing from a branch.
-  git fetch
-  if [[ "$(git rev-parse HEAD 2>&1)" != "$(git rev-parse '@{u}' 2>&1)" ]]; then
-    echo "There are upstream commits that won't be included in this release."
-    echo "You probably want to exit, run 'git pull', then release."
-    echo
-    read -p "Release anyway? " -n 1 -r
-    echo
-    if [[ $REPLY =~ ^[^Yy]$ ]]; then
-        echo Aborting.
-        exit 1
-    fi
+if [[ "$prerel" == "prerel" && "$BRANCH" != "prerelease" ]]; then
+  echo "❌ Sorry, you can only cut a pre-release from the 'prelease' branch"
+  echo "Run 'git checkout prerelease && git pull origin prerelease' and try again."
+  exit 1
+fi
+
+if [[ "$prerel" != "prerel" && "$BRANCH" != "master" && "$BRANCH" != "main" ]]; then
+  echo "❌ Sorry, you can only cut a release from the 'master' branch"
+  echo "Run 'git checkout master && git pull origin master' and try again."
+  exit 1
+fi
+
+git fetch
+if [[ "$(git rev-parse HEAD 2>&1)" != "$(git rev-parse '@{u}' 2>&1)" ]]; then
+  echo "There are upstream commits that won't be included in this release."
+  echo "You probably want to exit, run 'git pull', then release."
+  echo
+  read -p "Release anyway? " -n 1 -r
+  echo
+  if [[ $REPLY =~ ^[^Yy]$ ]]; then
+    echo Aborting.
+    exit 1
   fi
 fi
 
